#!/bin/bash

#
# Driver Installation Script
#
# This script should be run once to compile the dscudkp.o kernel module The
# kernel module will only load correctly under the kernel it was compiled for.
#
# This is required for driver interrupt operations.  If you are not using
# interrupt features of the driver there is no need to run this script.
#
# See the online documentation at http://docs.diamondsystems.com/dscud/
# for more information.
#

function check_linux_source() {

    if [ ! -e $1/COPYING ]; then
        return 1;
    fi

    if [ ! -e $1/Makefile ]; then
        return 1;
    fi

    if [ ! -e $1/.config ]; then
        return 1;
    fi

    EXTRAVERSION=
    eval `head $1/Makefile | sed -n 's/^\([A-Z]*\)[ ]*=[ ]*\([0-9a-zA-Z\.]*\)$/\1=\2/p'`
    if [ -z "$VERSION" -o -z "$PATCHLEVEL" -o -z "$SUBLEVEL" ]
    then
        return 1;
    fi

    return 0;
}

function get_source_version() {
    EXTRAVERSION=
    VERSION=
    PATCHLEVEL=
    SUBLEVEL=
    eval `head $1/Makefile | sed -n 's/^\([A-Z]*\)[ ]*=[ ]*\([0-9a-zA-Z\.]*\)$/\1=\2/p'`
    if [ -z "$VERSION" -o -z "$PATCHLEVEL" -o -z "$SUBLEVEL" ]
    then
        return 1;
    fi

    return 0;
}

echo
echo "Diamond Systems Universal Driver Installer"
echo "------------------------------------------"
echo
echo "This installer compiles a kernel module which supports"
echo "interrupt driven data acquisition features of Diamond "
echo "Systems products."
echo
echo "If you are not using the interrupt driven features of"
echo "the board you do not need to run this installer."
echo
echo "Online documentation for this driver is available at "
echo "http://docs.diamondsystems.com/dscud/"
echo
echo "Hit CONTROL-C at any time to exit this installer."
echo
echo "[[ HIT ENTER TO CONTINUE ]]"
read

echo "Step One: Locate Kernel Source Code"
echo "-------------------------------------------"
echo 
echo "The installer will now scan your system looking for"
echo "Linux kernel source code.  The /usr/src directory and"
echo "the /lib/modules directory will be scanned."
echo
echo "Select a Linux kernel version below that you will run"
echo "on your TARGET system.  Type the number next to the"
echo "selection and hit ENTER."
echo

for d in `ls -d1 /usr/src/* /lib/modules/*/build`
do
	check_linux_source $d
	if [ $? -eq 0 ]; then
		linux_src[${#linux_src[*]}]=$d
	fi
done

if [ ${#linux_src[@]} -eq 0 ]; then
	echo "The installer could not find any usable Linux source code."
	echo "See the online documentation above for help with this step."
	exit 1
fi

for i in $(seq 0 $((${#linux_src[@]} - 1)))
do
    dir=${linux_src[$i]};
    get_source_version $dir;
    echo "$i ) Kernel $VERSION.$PATCHLEVEL.$SUBLEVEL$EXTRAVERSION ($dir)";
done

read kversion
if [ X$kversion == "X" ]; then
	echo "No selection made.  Installer exiting."
	exit 1
fi

KDIR=${linux_src[$kversion]}

check_linux_source $KDIR
if [ $? -ne 0 ]; then
	echo "Invalid selection: $KDIR"
	exit 1
fi

if [ $PATCHLEVEL -eq 6 ]; then
	KERNEL=LINUX_26
else
	KERNEL=LINUX_24
fi

echo
echo "Step Two: Compile Kernel Module $modname"
echo "------------------------------------------"
echo
echo "The installer is now ready to compile the Linux kernel"
echo "module.  If errors occur see the README file in this"
echo "directory as well as the online DSCUD documentation"
echo "for help."
echo
echo "This kernel module will only load under the exact Linux"
echo "kernel version which you have installed in the directory"
echo "$KDIR."
echo

echo "[[ HIT ENTER TO CONTINUE ]]"
read

echo "--> Compiling kernel module for your system <--"
make KDIR=${KDIR} KERNEL=${KERNEL} clean
make KDIR=${KDIR} KERNEL=${KERNEL}
if [ $? -ne 0 ]; then
echo
echo
echo "An error occured while compiling the kernel module."
echo "See the README file in this directory and the online"
echo "DSCUD documentation for help."
exit
fi

echo "--> Installing module dscudkp.o in /lib/modules/misc <--"
make KDIR=${KDIR} KERNEL=${KERNEL} install

if [ $? -ne 0 ]; then
	echo
	echo
	echo "Kernel Compile Failed"
	echo "---------------------"
	echo
	echo "Review the README and online driver documentation"
	echo "for assistance with this step.  If you are unable"
	echo "to find an answer contact support@diamondsystems.com"
	echo "for help.  Be sure to include the exact error messages"
	echo "above in your e-mail."
	exit 1
fi

echo
echo "Step Three: Final Instructions"
echo "-----------------------------"
echo
echo "The dscudkp kernel module has been installed in"
echo "/lib/modules/misc/$modname.  You must copy this"
echo "file to the same location on your target system."
echo
echo "The load.sh script will load the kernel module so"
echo "that it can be used by the driver.  You must run"
echo "this script each time the Linux system boots.  See"
echo "the README file for help with this."
echo
echo "Driver installation complete."
echo
